#ifndef HTTP_METHOD
#define HTTP_METHOD(code, name)
#endif

HTTP_METHOD(0,  DELETE) \
HTTP_METHOD(1,  GET) \
HTTP_METHOD(2,  HEAD) \
HTTP_METHOD(3,  POST) \
HTTP_METHOD(4,  PUT) \
HTTP_METHOD(5,  CONNECT) \
HTTP_METHOD(6,  OPTIONS) \
HTTP_METHOD(7,  TRACE) \
HTTP_METHOD(8,  COPY) \
HTTP_METHOD(9,  LOCK) \
HTTP_METHOD(10, MKCOL) \
HTTP_METHOD(11, MOVE) \
HTTP_METHOD(12, PROPFIND) \
HTTP_METHOD(13, PROPPATCH) \
HTTP_METHOD(14, SEARCH) \
HTTP_METHOD(15, UNLOCK) \
HTTP_METHOD(16, BIND) \
HTTP_METHOD(17, REBIND) \
HTTP_METHOD(18, UNBIND) \
HTTP_METHOD(19, ACL) \
HTTP_METHOD(20, REPORT) \
HTTP_METHOD(21, MKACTIVITY) \
HTTP_METHOD(22, CHECKOUT) \
HTTP_METHOD(23, MERGE) \
HTTP_METHOD(24, M-SEARCH) \
HTTP_METHOD(25, NOTIFY) \
HTTP_METHOD(26, SUBSCRIBE) \
HTTP_METHOD(27, UNSUBSCRIBE) \
HTTP_METHOD(28, PATCH) \
HTTP_METHOD(29, PURGE) \
HTTP_METHOD(30, MKCALENDAR) \
HTTP_METHOD(31, LINK) \
HTTP_METHOD(32, UNLINK) \
HTTP_METHOD(33, SOURCE) \
HTTP_METHOD(34, PRI) \
HTTP_METHOD(35, DESCRIBE) \
HTTP_METHOD(36, ANNOUNCE) \
HTTP_METHOD(37, SETUP) \
HTTP_METHOD(38, PLAY) \
HTTP_METHOD(39, PAUSE) \
HTTP_METHOD(40, TEARDOWN) \
HTTP_METHOD(41, GET_PARAMETER) \
HTTP_METHOD(42, SET_PARAMETER) \
HTTP_METHOD(43, REDIRECT) \
HTTP_METHOD(44, RECORD) \
HTTP_METHOD(45, FLUSH) \

#undef HTTP_METHOD
